/*
 * Copyright (C) 2011 Koji Suzuki
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 2.1 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

`include "rtavr_defs.v"
// global defines : USE_DMY_CLOCK, OUTPUT_WB_ADDR, CHECK_CONFLICT

//`define USE_DMY_CLOCK
//`define OUTPUT_WB_ADDR
`define FOWARD_DI
//`define CHECK_CONFLICT

// GPR modele for the Reduced Core TinyAVR (attiny10) 
// version 005
// 
// Target Device: xc3s200a-4ft256
// Product Version: ISE 12.4
// Design Goal: Balanced
// LATCH_DOL : Insert LATCHes for Timing
//                                         2X
// Number of Slice Flip Flops:              67          51
// Number of 4 input LUTs:                 112         112
// Number of occupied Slices:               76          96
// Total Number of 4 input LUTs:           113         113
// Number used for Dual Port RAMs:          16          16
// Number of bonded IOBs:                   96          90
// Post-PAR Static Timing Report
// Clock to Setup (CLK2X) MAX            7.468              
//
//
module rtavr_gpr_16(
		input RESET,
`ifdef USE_DMY_CLOCK
		input CLK2X_IN,
		output CLK_OUT,
`else
		input CLK2X,
		input CLK,
`endif
		input WE,
		input PREDEC, POSTINC, 
		input [7:0] DI,
		input [3:0] ADDRAL,
		input [3:0] ADDRAH,
		input [3:0] ADDRBL,
		input [3:0] ADDRBH,
		output [7:0] DOBL,
		output [7:0] DOBH
`ifdef CHECK_CONFLICT
		, output DI_CONFLICT
`endif
		, output [11:0] JA
		, output [15:0] INDEX
`ifdef OUTPUT_WB_ADDR
		, output [3:0] WB_ADDR
`endif
		);
// Timing MEMO:
//
//                   _______________             ______________
// CLK2X ____________|              |____________|             |
//                                  ___________________________
// CLK  ____________________________|                          |
//
//
//      |            ------  ALL input --------                 |
//    ---- old value  ---           |  JA/INDEX low byte        |
//                                  |---|  JA       high byte   |
//                                  |----------|INDEX high byte |
//      |                      DOBL(old value)                  | DOBL
//      |                      DOBH(old value)                  | DOBH

`ifdef USE_DMY_CLOCK
  reg CLK;
  wire CLK2X;
  assign CLK_OUT = CLK;
  assign CLK2X   = CLK2X_IN;
`endif
  reg  [7:0] gpr [0:15];

  reg  [7:0] s2_DOBL;
  reg  [7:0] s2_DOBH;
  assign DOBL = s2_DOBL;
  assign DOBH = s2_DOBH;

  reg  [3:0] r_addrbl0;
  reg  [3:0] r_addrbh0;
  reg  [3:0] r_wb_addr0;

  reg  [3:0] r_addral;
  reg  [3:0] r_addrah;
  reg  [3:0] r_addrbl;
  reg  [3:0] r_addrbh;

  wire [15:0] DOA = { gpr[r_addrah] , gpr[r_addral] };

  wire WB = (PREDEC | POSTINC);

  reg  [3:0] r_wb_addr;
  `ifdef OUTPUT_WB_ADDR
    assign WB_ADDR = r_wb_addr;
  `endif
 
`ifdef GPR_WB_TUNE
  wire [15:0] DOA2 = (PREDEC) ? DOA - 16'h0001 : DOA;
  wire [15:0] WB_DATA = (PREDEC) ?  DOA - 16'h0001
                     : (POSTINC) ?  DOA + 16'h0001
		     : DOA;
`else
  wire [15:0] DOA2 = DOA - {15'h0000, PREDEC};
  wire [15:0] WB_DATA = DOA2 + {15'h0000, POSTINC};
`endif
  
  wire WE2L = WB ;
  wire WE2H = WB | WE ;
  wire [7:0] DI2L = WB_DATA[7:0];
  wire [7:0] DI2H = (WB) ? WB_DATA[15:8] : DI ;

`ifdef FOWARD_DI
  wire FORWARD_TO_BL = (WE) & (r_wb_addr == r_addrbl);
  wire FORWARD_TO_BH = (WE) & (r_wb_addr == r_addrbh);

  wire [7:0] v_DOBL = FORWARD_TO_BL ? DI2H : gpr[r_addrbl];
  wire [7:0] v_DOBH = FORWARD_TO_BH ? DI2H : gpr[r_addrbh];
`else
  wire [7:0] v_DOBL = gpr[r_addrbl];
  wire [7:0] v_DOBH = gpr[r_addrbh];
`endif

  assign JA[7:0] = v_DOBL;
  assign JA[11:8] = v_DOBH[3:0];
  assign INDEX[7:0] = DOA2[7:0];
`ifdef NO_STALL_IDX
  assign INDEX[15:8] = (WB) ? DOA2[15:8] : v_DOBH ;
`else
  assign INDEX[15:8] = DOA2[15:8];
`endif

`ifdef USE_DMY_CLOCK
always @(negedge CLK2X)
  begin
	CLK <= ~CLK;
  end

initial
  begin
	CLK <= 0;
  end
`endif

always @(posedge CLK)
  begin
    if (RESET == 0)
      begin
	r_addrbl0 <= 0;
	r_addrbh0 <= 0;
	r_wb_addr0 <= 0;
      end
    else
      begin
	r_addrbl0 <= ADDRBL;
	r_addrbh0 <= ADDRBH;
	if (WE2L) gpr[r_addral] <= DI2L;
	if (WE2H) gpr[r_addrah] <= DI2H;
`ifdef NO_STALL_IDX
	r_wb_addr0 <= (WE) ? ADDRAH : ADDRBH;
`else
	r_wb_addr0 <= (WE) ? ADDRAH : ADDRAL + 1;
`endif
      end
  end

  always @(negedge CLK)
  begin
    if (RESET == 0)
      begin

	r_addral <= 0;
	r_addrah <= 0;

	s2_DOBL <= 0;
	s2_DOBH <= 0;

	r_addrbl <= 0;
	r_addrbh <= 0;

	r_wb_addr <= 0;
      end
    else
      begin

	r_addral <= ADDRAL;
	r_addrah <= r_wb_addr;

	s2_DOBL <= v_DOBL;
	s2_DOBH <= v_DOBH;

//	r_addrbl <= ADDRBL;
//	r_addrbh <= ADDRBH;
	r_addrbl <= r_addrbl0;
	r_addrbh <= r_addrbh0; 
	r_wb_addr <= r_wb_addr0;
      end
  end

endmodule
